@@ -1,4 +1,4 @@
 from .main import run, Effect, Strand, TapystryError
 
-from .main import Send, Receive, CallFork, First, Call, Cancel
+from .main import Send, Receive, CallFork, First, Call, Cancel, Sleep
 from .effects import Fork, Join, Race, Subscribe
